package com.bergerkiller.bukkit.nolagg;

import net.minecraft.server.*;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

// CraftBukkit start
import org.bukkit.Bukkit;
import org.bukkit.entity.TNTPrimed;
import org.bukkit.event.entity.EntityDamageByBlockEvent;
import org.bukkit.event.entity.EntityDamageByEntityEvent;
import org.bukkit.event.entity.EntityDamageEvent;
import org.bukkit.event.entity.EntityExplodeEvent;
import org.bukkit.Location;
// CraftBukkit end

public class CustomExplosion {

    public boolean fire;
    private Random h = new Random();
    private World world;
    public double posX;
    public double posY;
    public double posZ;
    public Entity source;
    public float size;
    public Set<ChunkPosition> blocks = new HashSet<ChunkPosition>();

    public boolean wasCanceled = false; // CraftBukkit

    public CustomExplosion(World world, Entity entity, double x, double y, double z, float size, boolean fire) {
        this.world = world;
        this.source = entity;
        this.size = size;
        this.posX = x;
        this.posY = y;
        this.posZ = z;
        this.fire = fire;
    }
    
    public void doAll() {
    	this.doEntities();
    	this.doBlocks();
    }
           
    private static BlockChain[] chains;
    static {
    	chains = new BlockChain[1352];
    	int i = 0;
    	int x, y, z;
        for (x = 0; x < 16; ++x) {
            for (y = 0; y < 16; ++y) {
                for (z = 0; z < 16; ++z) {
                    if (x == 0 || x == 15 || y == 0 || y == 15 || z == 0 || z == 15) {
                    	chains[i] = new BlockChain((double) x, (double) y, (double) z);
                        i++;
                    }
                }
            }
        }
    }
    
    private static class BlockChain {
    	public BlockChain(double dx, double dy, double dz) {
        	dx = dx / 15.0 - 0.5;
        	dy = dy / 15.0 - 0.5;
        	dz = dz / 15.0 - 0.5;
        	double d = 0.2 / Math.sqrt(dx * dx + dy * dy + dz * dz);
        	this.dx = dx * d;
        	this.dy = dy * d;
        	this.dz = dz * d;
    	}
    	
    	public final double dx;
    	public final double dy;
    	public final double dz;
    	
    	private LinkedList<ExplosionBlock> blocks = new LinkedList<ExplosionBlock>();
    	
       	public void traverse(float damage, final World world, final int x, final int y, final int z, final Entity source) {
       		this.traverse(damage, world, new ChunkPosition(x, y, z), source);
       	}
    	public void traverse(float damage, final World world, final ChunkPosition pos, final Entity source) {
    		for (ExplosionBlock block : this.blocks) {
    			damage = block.traverse(damage, world, pos, source);
    			if (damage <= 0.0F) return;
    		}
    		//generate
    		while (damage > 0.0F) {
    			ExplosionBlock block = getBlock(this.dx, this.dy, this.dz, this.blocks.size());
    			this.blocks.offer(block);
    			damage = block.traverse(damage, world, pos, source);
    		}
    	}
    	
    }
    
    private static List<ExplosionBlock> explosionBlocks = new ArrayList<ExplosionBlock>();
    private static Map<ChunkPosition, ExplosionBlock> explosionBlockMap = new HashMap<ChunkPosition, ExplosionBlock>();
    
    
    private static ExplosionBlock getBlock(double dx, double dy, double dz, int factor) {
    	return getBlock(dx, dy, dz, (double) factor);
    }
    private static ExplosionBlock getBlock(double dx, double dy, double dz, double factor) {
    	return getBlock(dx * factor, dy * factor, dz * factor);
    }
    private static ExplosionBlock getBlock(double dx, double dy, double dz) {
    	return getBlock(new ChunkPosition(MathHelper.floor(dx), MathHelper.floor(dy), MathHelper.floor(dz)));
    }
    private static ExplosionBlock getBlock(ChunkPosition pos) {
    	ExplosionBlock block = explosionBlockMap.get(pos);
    	if (block == null) {
    		block = new ExplosionBlock(pos);
    		explosionBlocks.add(block);
    		explosionBlockMap.put(pos, block);
    	}
    	return block;
    }
    
    private static class ExplosionBlock {
    	public ExplosionBlock(final ChunkPosition pos) {
    		this.pos = pos;
    	}
    	public final ChunkPosition pos;
  
    	public boolean isSet = false;
    	public boolean destroy = false;
    	public int type;
    	public float damagefactor;
    	
    	//traversing the slots
    	public float traverse(float damage, final World world, final ChunkPosition pos, final Entity source) {
    		if (!this.isSet) {
        		this.type = world.getTypeId(pos.x + this.pos.x, pos.y + this.pos.y, pos.z + this.pos.z);
        		if (this.type > 0) {
        			this.damagefactor = (Block.byId[type].a(source) + 0.3F) * 0.3F;
        		} else {
        			this.damagefactor = 0;
        		}
    			this.isSet = true;
    		}
    		damage -= this.damagefactor;
    		if (damage > 0.0F) this.destroy = true;
    		damage -= 0.225;
    		return damage;
    	}
    }

    public void doEntities() {
        float f = this.size;

        int x = MathHelper.floor(this.posX);
        int y = MathHelper.floor(this.posY);
        int z = MathHelper.floor(this.posZ);

        for (BlockChain chain : chains) {
        	chain.traverse(this.size * (0.7F + this.world.random.nextFloat() * 0.6F), world, x, y, z, source);
        }
        
        for (ExplosionBlock block : explosionBlocks) {
        	if (block.destroy) {
        		this.blocks.add(new ChunkPosition(block.pos.x + x, block.pos.y + y, block.pos.z + z));
            	block.destroy = false;
        	}
        	block.isSet = false;
        }

        double d0;
        double d1;
        double d2;

        this.size *= 2.0F;
        x = MathHelper.floor(this.posX - (double) this.size - 1.0);
        y = MathHelper.floor(this.posX + (double) this.size + 1.0);
        z = MathHelper.floor(this.posY - (double) this.size - 1.0);
        int l1 = MathHelper.floor(this.posY + (double) this.size + 1.0D);
        int i2 = MathHelper.floor(this.posZ - (double) this.size - 1.0D);
        int j2 = MathHelper.floor(this.posZ + (double) this.size + 1.0D);
        List list = this.world.getEntities(this.source, AxisAlignedBB.b((double) x, (double) z, (double) i2, (double) y, (double) l1, (double) j2));
        Vec3D vec3d = Vec3D.create(this.posX, this.posY, this.posZ);

        for (int k2 = 0; k2 < list.size(); ++k2) {
            Entity entity = (Entity) list.get(k2);
            double d7 = entity.f(this.posX, this.posY, this.posZ) / (double) this.size;

            if (d7 <= 1.0D) {
                d0 = entity.locX - this.posX;
                d1 = entity.locY - this.posY;
                d2 = entity.locZ - this.posZ;
                double d8 = Math.sqrt(d0 * d0 + d1 * d1 + d2 * d2); // CraftBukkit

                d0 /= d8;
                d1 /= d8;
                d2 /= d8;
                double d9 = (double) this.world.a(vec3d, entity.boundingBox);
                double d10 = (1.0D - d7) * d9;

                // CraftBukkit start - explosion damage hook
                org.bukkit.entity.Entity damagee = (entity == null) ? null : entity.getBukkitEntity();
                int damageDone = (int) ((d10 * d10 + d10) / 2.0D * 8.0D * (double) this.size + 1.0D);

                if (damagee == null) {
                    // nothing was hurt
                } else if (this.source == null) { // Block explosion
                    // TODO: get the x/y/z of the tnt block?
                    // does this even get called ever? @see EntityTNTPrimed - not BlockTNT or whatever
                    EntityDamageByBlockEvent event = new EntityDamageByBlockEvent(null, damagee, EntityDamageEvent.DamageCause.BLOCK_EXPLOSION, damageDone);
                    Bukkit.getPluginManager().callEvent(event);

                    if (!event.isCancelled()) {
                        entity.damageEntity(DamageSource.EXPLOSION, event.getDamage());
                        entity.motX += d0 * d10;
                        entity.motY += d1 * d10;
                        entity.motZ += d2 * d10;
                    }
                } else {
                    final org.bukkit.entity.Entity damager = this.source.getBukkitEntity();
                    final EntityDamageEvent.DamageCause damageCause;
                    if (damager instanceof TNTPrimed) {
                        damageCause = EntityDamageEvent.DamageCause.BLOCK_EXPLOSION;
                    } else {
                        damageCause = EntityDamageEvent.DamageCause.ENTITY_EXPLOSION;
                    }
                    EntityDamageByEntityEvent event = new EntityDamageByEntityEvent(damager, damagee, damageCause, damageDone);
                    Bukkit.getPluginManager().callEvent(event);

                    if (!event.isCancelled()) {
                        entity.damageEntity(DamageSource.EXPLOSION, event.getDamage());

                        entity.motX += d0 * d10;
                        entity.motY += d1 * d10;
                        entity.motZ += d2 * d10;
                    }
                }
                // CraftBukkit end
            }
        }

        this.size = f;
        ArrayList arraylist = new ArrayList();

        arraylist.addAll(this.blocks);
    }

    public void doBlocks() {
        //this.world.makeSound(this.posX, this.posY, this.posZ, "random.explode", 4.0F, (1.0F + (this.world.random.nextFloat() - this.world.random.nextFloat()) * 0.2F) * 0.7F);
        //this.world.a("hugeexplosion", this.posX, this.posY, this.posZ, 0.0D, 0.0D, 0.0D);
       
    	ArrayList<ChunkPosition> arraylist = new ArrayList<ChunkPosition>(this.blocks);

        // CraftBukkit start
        org.bukkit.World bworld = this.world.getWorld();
        org.bukkit.entity.Entity explode = this.source == null ? null : this.source.getBukkitEntity();
        Location location = new Location(bworld, this.posX, this.posY, this.posZ);

        List<org.bukkit.block.Block> blockList = new ArrayList<org.bukkit.block.Block>(arraylist.size());
        for (int j = arraylist.size() - 1; j >= 0; j--) {
            ChunkPosition cpos = (ChunkPosition) arraylist.get(j);
            org.bukkit.block.Block block = bworld.getBlockAt(cpos.x, cpos.y, cpos.z);
            if (block.getType() != org.bukkit.Material.AIR) {
                blockList.add(block);
            }
        }

        EntityExplodeEvent event = new EntityExplodeEvent(explode, location, blockList, 0.3F);
        Util.callEvent(event);

        arraylist.clear();
        this.blocks.clear();

        for (org.bukkit.block.Block block : event.blockList()) {
            ChunkPosition coords = new ChunkPosition(block.getX(), block.getY(), block.getZ());
            arraylist.add(coords);
            this.blocks.add(coords);
        }

        if (event.isCancelled()) {
            this.wasCanceled = true;
            return;
        }
        // CraftBukkit end

        int i;
        ChunkPosition chunkposition;
        int j;
        int k;
        int l;
        int i1;

        for (i = arraylist.size() - 1; i >= 0; --i) {
            chunkposition = (ChunkPosition) arraylist.get(i);
            j = chunkposition.x;
            k = chunkposition.y;
            l = chunkposition.z;
            i1 = this.world.getTypeId(j, k, l);
            
            double d0 = (double) ((float) j + this.world.random.nextFloat());
            double d1 = (double) ((float) k + this.world.random.nextFloat());
            double d2 = (double) ((float) l + this.world.random.nextFloat());
            double d3 = d0 - this.posX;
            double d4 = d1 - this.posY;
            double d5 = d2 - this.posZ;
            double d6 = (double) MathHelper.sqrt(d3 * d3 + d4 * d4 + d5 * d5);

            d3 /= d6;
            d4 /= d6;
            d5 /= d6;
            double d7 = 0.5D / (d6 / (double) this.size + 0.1D);

            d7 *= (double) (this.world.random.nextFloat() * this.world.random.nextFloat() + 0.3F);
            d3 *= d7;
            d4 *= d7;
            d5 *= d7;
            
            //this.world.a("explode", (d0 + this.posX * 1.0D) / 2.0D, (d1 + this.posY * 1.0D) / 2.0D, (d2 + this.posZ * 1.0D) / 2.0D, d3, d4, d5);
            //this.world.a("smoke", d0, d1, d2, d3, d4, d5);

            // CraftBukkit - stop explosions from putting out fire
            if (i1 > 0 && i1 != Block.FIRE.id) {
                // CraftBukkit
                Block.byId[i1].dropNaturally(this.world, j, k, l, this.world.getData(j, k, l), event.getYield(), 0);
                this.world.setTypeId(j, k, l, 0);
                Block.byId[i1].wasExploded(this.world, j, k, l);
            }
        }

        if (this.fire) {
            for (i = arraylist.size() - 1; i >= 0; --i) {
                chunkposition = (ChunkPosition) arraylist.get(i);
                j = chunkposition.x;
                k = chunkposition.y;
                l = chunkposition.z;
                i1 = this.world.getTypeId(j, k, l);
                int j1 = this.world.getTypeId(j, k - 1, l);

                if (i1 == 0 && Block.o[j1] && this.h.nextInt(3) == 0) {
                    this.world.setTypeId(j, k, l, Block.FIRE.id);
                }
            }
        }
    }
}
